package polyglot.ext.ibex.lr;

import java.util.*;
import polyglot.main.Report;
import polyglot.util.InternalCompilerError;

/**
 * This class contains code that converts a context free grammar into
 * a DFA for a regular approximation of the language.
 *
 * The basic translation is due to Nederhof and is decribed in the paper
 * "Practical Experiments with Regular pproximation" in Computational
 * Linguistics, 26(1).  Section 4.1.
 */
class Partition
{
    public static final int NONE = 0;
    public static final int LEFT = 1;
    public static final int RIGHT = 2;
    public static final int CYCLIC = 3;
    public static final int SELF = 4;

    List nonterminals;

    BitSet leftGenerating;
    BitSet rightGenerating;
    BitSet recursive;

    UnionFind partitions;
    Collection[] partition;

    /**
     * Build a regular approximation for the rule's RHS.
     * The LHS must be a augmented grammar start symbol; it cannot
     * be reachable from the RHS.
     */
    Partition(List nonterminals, GLRNonterminal S) {
        this.nonterminals = nonterminals;

        if (S.rules().size() != 1) {
            throw new InternalCompilerError(
                "Cannot create NFA for non-augmented grammar.");
        }

        this.partition = new Collection[nonterminals.size()];
        this.partitions = new UnionFind();
        this.recursive = new BitSet();
        findPartitions(S);

        // Now for each partition, determine if the partition is left-
        // or right- generating.  Using this information, partitions can
        // be classified as left, right, cyclic, or self recursive.
        this.leftGenerating = new BitSet();
        this.rightGenerating = new BitSet();
        computeLeftRightGeneratingDFS(S, new BitSet());

        System.out.println("partitions for " + S);
        for (int i = 0; i < partition.length; i++) {
            if (partition[i] != null) {
                System.out.print(i + ": " + partition[i] + " ");
                switch (recursive(i)) {
                    case NONE: System.out.println("non-recursive"); break;
                    case LEFT: System.out.println("left"); break;
                    case RIGHT: System.out.println("right"); break;
                    case CYCLIC: System.out.println("cyclic"); break;
                    case SELF: System.out.println("self"); break;
                }
            }
        }
    }

    public Collection members(GLRNonterminal A) {
        return partition[partition(A)];
    }

    private void collectPostorderDFS(GLRNonterminal A, List postorder, List[] preds, BitSet visited) {
        if (visited.get(A.index())) {
            return;
        }

        visited.set(A.index());

        for (Iterator j = A.rules().iterator(); j.hasNext(); ) {
            GLRRule rule = (GLRRule) j.next();

            for (Iterator i = rule.rhs.iterator(); i.hasNext(); ) {
                GLRSymbol X = (GLRSymbol) i.next();

                if (X instanceof GLRNonterminal) {
                    GLRNonterminal B = (GLRNonterminal) X;

                    if (preds[B.index()] == null) {
                        preds[B.index()] = new ArrayList();
                    }

                    preds[B.index()].add(A);

                    collectPostorderDFS(B, postorder, preds, visited);
                }
            }
        }

        postorder.add(A);
    }

    // Partition the recursive nonterminals into sets of mutually
    // recursive nonterminals.
    // This is done by finding SCCs for the graph of nonterminals
    // where there is an edge from A to B if B appears on the RHS
    // of a rule for A.
    private void findPartitions(GLRNonterminal S) {
        // First collect all the nonterminals in post order.
        List postorder = new ArrayList();
        List[] preds = new List[nonterminals.size()];
        collectPostorderDFS(S, postorder, preds, new BitSet());

        System.out.println("postorder = " + postorder);

        // Now, walk the transposed graph picking nodes in reverse
        // postorder, thus picking out one SCC at a time and
        // appending it to "by_scc".
        int n = postorder.size();
        BitSet visited = new BitSet();

        Iterator[] p = new Iterator[nonterminals.size()];

	for (int i = 0; i < preds.length; i++) {
            if (preds[i] != null) {
                p[i] = preds[i].iterator();
            }
            else {
                p[i] = Collections.EMPTY_LIST.iterator();
            }
        }

        // Go in reverse postorder
	for (int i = n-1; i >= 0; i--) {
            GLRNonterminal A = (GLRNonterminal) postorder.get(i);

	    if (! visited.get(A.index())) {
		visited.set(A.index());

                GLRNonterminal[] stack = new GLRNonterminal[n];
                int top = 0;

                stack[top++] = A;

                while (top > 0) {
                    GLRNonterminal C = stack[top-1];

                    if (p[C.index()].hasNext()) {
                        GLRNonterminal B = (GLRNonterminal) p[C.index()].next();

                        if (! visited.get(B.index())) {
                            stack[top++] = B;
                        }
                    }
                    else {
                        top--;
                        if (C != A) {
                            System.out.println("union " + C + " and " + A);
                            partitions.union(C.index(), A.index());
                        }
                    }
                }
	    }
	}

	for (int i = 0; i < n; i++) {
            GLRNonterminal A = (GLRNonterminal) postorder.get(i);

            int Ni = partition(A);

            if (partition[Ni] == null) {
                partition[Ni] = new ArrayList();
            }

            partition[Ni].add(A);
        }

	for (int i = 0; i < n; i++) {
            GLRNonterminal A = (GLRNonterminal) postorder.get(i);
            int Ni = partition(A);
            if (partition[Ni].size() > 1) {
                recursive.set(A.index());
            }
            else if (preds[A.index()] != null) {
                // Check if there's a rule A ->* A
                for (Iterator j = preds[A.index()].iterator(); j.hasNext(); ) {
                    GLRNonterminal B = (GLRNonterminal) j.next();
                    if (A == B) {
                        recursive.set(A.index());
                    }
                }
            }
        }
    }


    // Determine which partitions are left generating and which are right
    // generating.
    private void computeLeftRightGeneratingDFS(GLRNonterminal A, BitSet visited) {
        if (visited.get(A.index())) {
            return;
        }

        visited.set(A.index());

        int a = partition(A);

        for (Iterator j = A.rules().iterator(); j.hasNext(); ) {
            GLRRule rule = (GLRRule) j.next();

            boolean first = true;

            for (Iterator i = rule.rhs.iterator(); i.hasNext(); ) {
                GLRSymbol X = (GLRSymbol) i.next();

                boolean last = ! i.hasNext();

                if (X instanceof GLRNonterminal) {
                    GLRNonterminal B = (GLRNonterminal) X;
                    computeLeftRightGeneratingDFS(B, visited);

                    int b = partition(B);

                    if (a == b) {
                        // Important: use "if", not "else if".  A
                        // nonterminal can be both left and right
                        // generating if the rule is a unit production
                        // A -> B
                        if (! first) {
                            leftGenerating.set(a);
                        }
                        if (! last) {
                            rightGenerating.set(a);
                        }
                    }
                }

                first = false;
            }
        }
    }

    public boolean contains(GLRNonterminal A, GLRSymbol X) {
        int Ni = partition(A);
        return contains(Ni, X);
    }

    private boolean contains(int Ni, GLRSymbol X) {
        if (X instanceof GLRNonterminal) {
            int x = partition((GLRNonterminal) X);
            return x == Ni;
        }
        return false;
    }

    public boolean containsAny(GLRNonterminal A, List rhs) {
        int Ni = partition(A);
        for (int i = 0; i < rhs.size(); i++) {
            GLRSymbol X = (GLRSymbol) rhs.get(i);
            if (contains(Ni, X)) {
                return true;
            }
        }
        return false;
    }

    public boolean same(GLRNonterminal A, GLRNonterminal B) {
        return partition(A) == partition(B);
    }

    public int partition(GLRNonterminal A) {
        return partitions.find(A.index());
    }

    public int recursive(GLRNonterminal A) {
        return recursive(partition(A));
    }

    private int recursive(int Ni) {
        if (recursive.get(Ni)) {
            if (! leftGenerating.get(Ni) && rightGenerating.get(Ni)) {
                return LEFT;
            }
            if (leftGenerating.get(Ni) && ! rightGenerating.get(Ni)) {
                return RIGHT;
            }
            if (leftGenerating.get(Ni) && rightGenerating.get(Ni)) {
                return SELF;
            }
            if (! leftGenerating.get(Ni) && ! rightGenerating.get(Ni)) {
                return CYCLIC;
            }
        }
        return NONE;
    }
}
