package polyglot.ext.ibex.visit;

import polyglot.ast.*;
import polyglot.ext.ibex.ast.*;
import polyglot.ext.ibex.types.*;
import polyglot.ext.ibex.lr.GLR;
import polyglot.ext.ibex.ExtensionInfo;
import polyglot.frontend.*;
import polyglot.types.*;
import polyglot.util.*;
import polyglot.visit.*;
import polyglot.types.Package;
import polyglot.main.Report;

import java.io.IOException;
import java.util.*;

/** Visitor which traverses the AST constructing type objects. */
public class Rewriter extends ContextVisitor
{
    GLR glr;

    public Rewriter(Job job, TypeSystem ts, NodeFactory nf) {
        super(job, ts, nf);
        glr = null;
    }

    public IbexTypeSystem glrTypeSystem() {
        return (IbexTypeSystem) typeSystem();
    }

    public IbexNodeFactory glrNodeFactory() {
        return (IbexNodeFactory) nodeFactory();
    }

    protected NodeVisitor enterCall(Node parent, Node n) throws SemanticException {
        if (n instanceof ClassBody) {
            this.glr = null;
        }

        return super.enterCall(parent, n);
    }

    Map glrs = new HashMap();
    Map membersMap = new HashMap();

    GLR glr(ParserType pt) {
        if (glr == null) {
            glr = (GLR) glrs.get(pt);
            if (glr == null) {
                glr = new GLR((ExtensionInfo) job().extensionInfo(), pt);
                glrs.put(pt, glr);
            }
        }
        return glr;
    }

    List members(ParserType pt) {
        List l = (List) membersMap.get(pt);
        if (l == null) {
            l = new ArrayList();
            membersMap.put(pt, l);
        }
        return l;
    }

    protected Node leaveCall(Node old, Node n, NodeVisitor v) throws SemanticException {
        // Use the context used to visit children to get the current class.
        ContextVisitor v_ = (ContextVisitor) v;
        Context c = v_.context();

        if (n instanceof ClassDecl) {
            ClassDecl cd = (ClassDecl) n;
            MutableParserType pt = (MutableParserType) cd.type();

            IbexTypeSystem ts = (IbexTypeSystem) pt.typeSystem();

            // Ensure the class implements ibex.runtime.ParserImpl
            // and not just ibex.runtime.Parser
            ClassType impl = ts.runtimeParserImplType();

            if (pt.isParser() && ! pt.isSubtype(impl)) {
                pt.addInterface(impl);
                List interfaces = new ArrayList(cd.interfaces());
                interfaces.add(nf.CanonicalTypeNode(n.position(), impl));
                return cd.interfaces(interfaces);
            }
        }
        else if (n instanceof ClassBody) {
            ParserType pt = (ParserType) c.currentClass();

            ClassBody cb = (ClassBody) n;

            if (pt.isParser()) {
                MutableParserType mpt =
                    (MutableParserType) c.currentClassScope();
                addGlrTableMembers(mpt, glr(pt));
                addGlrEof(mpt, glr(pt));
                addGlrSemanticActions(mpt, glr(pt));
                addGlrMergeFunctions(mpt, glr(pt));
            }

            glr = null;
            return cb.members(members(pt));
        }
        else if (n instanceof IbexClassMember) {
            ParserType pt = (ParserType) c.currentClass();
            glr = glr(pt);
            List l = ((IbexClassMember) n).rewrite(this);
            members(pt).addAll(l);
        }
        else if (n instanceof ClassMember) {
            ParserType pt = (ParserType) c.currentClass();
            members(pt).add(n);
        }

        return super.leaveCall(old, n, v);
    }

    public GLR glr() {
        if (glr == null) {
            throw new InternalCompilerError("GLR object not initialized.");
        }
        return glr;
    }

    protected void addGlrMergeFunctions(MutableParserType pt, GLR glr)
        throws SemanticException
    {
        IbexTypeSystem ts = (IbexTypeSystem) pt.typeSystem();

        // add method:
        // Object merge(int symbol, Object o1, Object o2) {
        //     switch (symbol) {
        //         case 0:
        //             return marshal(merge$N0(unmarshal(o1), unmarshal(o1)));
        //         ...
        //     }
        //     return ((Mergeable) o1).merge(o2);
        // }

        Position pos = Position.COMPILER_GENERATED;

        Type objectArray = ts.arrayOf(pos, ts.Object());

        ClassType mergeable = ts.runtimeMergeableType();
        TypeNode mtn = nf.CanonicalTypeNode(pos, mergeable);
        TypeNode otn = nf.CanonicalTypeNode(pos, ts.Object());
        TypeNode atn = nf.CanonicalTypeNode(pos, objectArray);
        TypeNode itn = nf.CanonicalTypeNode(pos, ts.Int());

        MethodInstance mmi;

        try {
            mmi = ts.findMethod(mergeable, "merge",
                                Collections.singletonList(mergeable), pt);
        }
        catch (SemanticException e) {
            throw new InternalCompilerError(e);
        }

        LocalInstance rule1LI = ts.localInstance(pos, Flags.NONE,
                                              ts.Int(), "rule1");
        LocalInstance rule2LI = ts.localInstance(pos, Flags.NONE,
                                              ts.Int(), "rule2");
        LocalInstance sval1LI = ts.localInstance(pos, Flags.NONE,
                                              ts.arrayOf(pos, ts.Object()),
                                              "sval1");
        LocalInstance sval2LI = ts.localInstance(pos, Flags.NONE,
                                              ts.arrayOf(pos, ts.Object()),
                                              "sval2");

        Local rule1 = nf.Local(pos, rule1LI.name());
        rule1 = rule1.localInstance(rule1LI);
        rule1 = (Local) rule1.type(rule1LI.type());

        Local rule2 = nf.Local(pos, rule2LI.name());
        rule2 = rule2.localInstance(rule2LI);
        rule2 = (Local) rule2.type(rule2LI.type());

        Local sval1 = nf.Local(pos, sval1LI.name());
        sval1 = sval1.localInstance(sval1LI);
        sval1 = (Local) sval1.type(sval1LI.type());

        Local sval2 = nf.Local(pos, sval2LI.name());
        sval2 = sval2.localInstance(sval2LI);
        sval2 = (Local) sval2.type(sval2LI.type());

        List elements = new ArrayList(2*glr.numRules());

        // build the case arms for nonterminals with explicit merge actions.
        for (Iterator i = pt.nonterminals().iterator(); i.hasNext(); ) {
            Nonterminal lhs = (Nonterminal) i.next();

            for (Iterator j = lhs.choices().iterator(); j.hasNext(); ) {
                Rhs choice = (Rhs) j.next();

                if (! (choice instanceof RhsMerge)) {
                    continue;
                }

                RhsMerge merge = (RhsMerge) choice;
                RhsCase rhs1 = merge.rhs1();
                RhsCase rhs2 = merge.rhs2();

                int ruleNumber1 = glr.ruleNumber(lhs, rhs1);
                int ruleNumber2 = glr.ruleNumber(lhs, rhs2);

                // The lower numbered rule should be first.
                if (ruleNumber2 <= ruleNumber1) {
                    int tmp = ruleNumber1;
                    ruleNumber1 = ruleNumber2;
                    ruleNumber2 = tmp;
                }

                int rule = (ruleNumber1 << 16) | ruleNumber2;

                // case i:
                IntLit label = nf.IntLit(pos, IntLit.INT, rule);
                label = (IntLit) label.type(ts.Int());
                Case case_ = nf.Case(pos, label);
                case_ = case_.value(rule);
                elements.add(case_);

                // build the call:
                // merge$lhs$rhs1$rhs2(unmarshal(rhs1), unmarshal(rhs2))
                List code = new ArrayList();

                String methodName = merge.actionMethodName(lhs);
                List argTypes = merge.actionMethodFormalTypes();

                MethodInstance mi = ts.methodInstance(
                    pos, pt, Flags.PROTECTED, lhs.symbolType(), methodName,
                    argTypes, Collections.EMPTY_LIST);

                List args = new ArrayList();
                unmarshalSvals(rhs1, sval1, code, args, pt);
                unmarshalSvals(rhs2, sval2, code, args, pt);

                if (args.size() != argTypes.size()) {
                    throw new InternalCompilerError("Argument count mismatch between " + mi + " and " + args);
                }

                Special this_ = nf.This(pos);
                this_ = (Special) this_.type(pt);
                Call call = nf.Call(pos, this_, mi.name(), args);
                call = call.methodInstance(mi);
                call = (Call) call.type(mi.returnType());

                // marshal the result and return it
                Return ret = nf.Return(pos, marshal(call, mi.returnType(), pt));
                code.add(ret);

                Block b = nf.Block(pos, code);
                SwitchBlock sb = nf.SwitchBlock(pos, Collections.singletonList(b));
                elements.add(sb);
            }
        }

        // build the switch statement.
        List stmts = new ArrayList(2);

        if (! elements.isEmpty()) {
            Expr sixteen = nf.IntLit(pos, IntLit.INT, 16);
            sixteen = sixteen.type(ts.Int());
            Expr shift = nf.Binary(pos, rule1, Binary.SHL, sixteen);
            shift = shift.type(ts.Int());
            Expr index = nf.Binary(pos, shift, Binary.BIT_OR, rule2);
            index = index.type(ts.Int());

            Switch s = nf.Switch(pos, index, elements);
            stmts.add(s);
        }

        // build the default fall-through case:
        //         return null;
        Expr null_ = nf.NullLit(pos);
        null_ = null_.type(ts.Null());
        Return retN = nf.Return(pos, null_);
        stmts.add(retN);

/*
        // build the default fall-through case:
        //     if (sval1 != null)
        //         return ((Mergeable) sval1).merge((Mergeable) sval2);
        //     else if (sval2 != null)
        //         return ((Mergeable) sval2).merge((Mergeable) sval1);
        //     else
        //         return null;
        Cast cast1 = nf.Cast(pos, mtn, o1);
        Cast cast2 = nf.Cast(pos, mtn, o2);

        Call call1 = nf.Call(pos, cast1, mmi.name(),
                            Collections.singletonList(cast2));
        call1 = call1.methodInstance(mmi);
        call1 = (Call) call1.type(mmi.returnType());

        Call call2 = nf.Call(pos, cast2, mmi.name(),
                            Collections.singletonList(cast1));
        call2 = call2.methodInstance(mmi);
        call2 = (Call) call2.type(mmi.returnType());

        Expr null_ = nf.NullLit(pos);
        null_ = null_.type(ts.Null());

        Return ret1 = nf.Return(pos, call1);
        Return ret2 = nf.Return(pos, call2);
        Return retN = nf.Return(pos, null_);

        Expr eqNull1 = nf.Binary(pos, o1, Binary.NE, null_);
        eqNull1 = eqNull1.type(ts.Boolean());
        Expr eqNull2 = nf.Binary(pos, o2, Binary.NE, null_);
        eqNull2 = eqNull2.type(ts.Boolean());

        If if2 = nf.If(pos, eqNull2, ret2, retN);
        If if1 = nf.If(pos, eqNull1, ret1, if2);

        stmts.add(if1);
*/

        // build the method body.
        Block b = nf.Block(pos, stmts);

        // now build the method declaration.
        Formal rule1f = nf.Formal(pos, Flags.NONE, itn, rule1LI.name());
        rule1f = rule1f.localInstance(rule1LI);
        Formal rule2f = nf.Formal(pos, Flags.NONE, itn, rule2LI.name());
        rule2f = rule2f.localInstance(rule2LI);
        Formal sval1f = nf.Formal(pos, Flags.NONE, atn, sval1LI.name());
        sval1f = sval1f.localInstance(sval1LI);
        Formal sval2f = nf.Formal(pos, Flags.NONE, atn, sval2LI.name());
        sval2f = sval2f.localInstance(sval2LI);

        List formals = new ArrayList(2);
        formals.add(rule1f);
        formals.add(rule2f);
        formals.add(sval1f);
        formals.add(sval2f);

        MethodInstance mi;

        List formalTypes = new ArrayList(2);
        formalTypes.add(rule1LI.type());
        formalTypes.add(rule2LI.type());
        formalTypes.add(sval1LI.type());
        formalTypes.add(sval2LI.type());

        try {
            mi = ts.findMethod(ts.runtimeParserImplType(), "mergeAction",
                               formalTypes, pt);
            mi = mi.flags(Flags.PUBLIC);
            mi = mi.container(pt);
        }
        catch (SemanticException e) {
            throw new InternalCompilerError(e);
        }

        MethodDecl m = nf.MethodDecl(pos, mi.flags(), otn, mi.name(),
                formals, Collections.EMPTY_LIST, b);
        m = m.methodInstance(mi);

        pt.addMethod(mi);

        members(pt).add(m);
    }

    protected void unmarshalSvals(RhsCase rhs, Expr sval, List code, List args, MutableParserType pt) {
        Position pos = Position.COMPILER_GENERATED;

        int k = 0;

        Iterator ei = rhs.elements().iterator();
        Iterator ti = rhs.tags().iterator();

        while (ei.hasNext() && ti.hasNext()) {
            Symbol s = (Symbol) ei.next();
            String tag = (String) ti.next();

            if (tag != null) {
                  // k: the index into the sval array
                  IntLit index = nf.IntLit(pos, IntLit.INT, k);
                  index = (IntLit) index.type(ts.Int());

                  // sval[k]
                  ArrayAccess access = nf.ArrayAccess(pos, sval, index);
                  access = (ArrayAccess) access.type(ts.Object());

                  LocalInstance li = ts.localInstance(pos, Flags.NONE,
                                                      s.symbolType(), tag);

                  // T tag = (T) sval[k]
                  LocalDecl d = nf.LocalDecl(pos, li.flags(),
                      nf.CanonicalTypeNode(pos, li.type()), li.name(),
                      unmarshal(access, li.type(), pt));
                  d = d.localInstance(li);

                  code.add(d);

                  Local l = nf.Local(pos, li.name());
                  l = l.localInstance(li);
                  l = (Local) l.type(li.type());

                  args.add(l);
            }

            k++;
        }
    }

    protected Expr unmarshal(Expr e, Type t, ParserType pt) {
        TypeSystem ts = pt.typeSystem();
        Position pos = Position.COMPILER_GENERATED;

        try {
            if (t.isReference()) {
                TypeNode tn = nf.CanonicalTypeNode(pos, t);
                Cast cast = nf.Cast(pos, tn, e);
                cast = (Cast) cast.type(t);
                return cast;
            }
            else if (t.isPrimitive()) {
                // ((Integer) e).intValue();

                // Get Integer
                String wrapperName = ts.wrapperTypeString(t.toPrimitive());
                ClassType wrapperType = (ClassType) ts.typeForName(wrapperName);
                TypeNode tn = nf.CanonicalTypeNode(pos, wrapperType);

                // (Integer) e
                Cast cast = nf.Cast(pos, tn, e);
                cast = (Cast) cast.type(wrapperType);

                // "intValue"
                String methodName = t.toPrimitive().name() + "Value";

                MethodInstance mi = ts.findMethod(wrapperType, methodName,
                                                  Collections.EMPTY_LIST, pt);

                // ((Integer) e).intValue()
                Call call = nf.Call(pos, cast, methodName,
                                    Collections.EMPTY_LIST);
                call = call.methodInstance(mi);
                call = (Call) call.type(mi.returnType());

                return call;
            }
        }
        catch (SemanticException ex) {
            throw new InternalCompilerError(ex);
        }

        throw new InternalCompilerError("Cannot unmarshal type " + t);
    }

    protected Expr marshal(Expr e, Type t, ParserType pt) {
        TypeSystem ts = pt.typeSystem();
        Position pos = Position.COMPILER_GENERATED;

        try {
            if (t.isReference()) {
                return e;
            }
            else if (t.isPrimitive()) {
                // new Integer(e);

                // Integer
                String wrapperName = ts.wrapperTypeString(t.toPrimitive());
                ClassType wrapperType = (ClassType) ts.typeForName(wrapperName);
                TypeNode tn = nf.CanonicalTypeNode(pos, wrapperType);

                ConstructorInstance ci = ts.findConstructor(wrapperType,
                    Collections.singletonList(t), pt);

                // new Integer(e)
                New new_ = nf.New(pos, tn, Collections.singletonList(e));
                new_ = (New) new_.constructorInstance(ci);
                new_ = (New) new_.type(wrapperType);

                return new_;
            }
        }
        catch (SemanticException ex) {
            throw new InternalCompilerError(ex);
        }

        throw new InternalCompilerError("Cannot marshal type " + t);
    }

    protected void addGlrSemanticActions(MutableParserType pt, GLR glr)
        throws SemanticException
    {
        IbexTypeSystem ts = (IbexTypeSystem) pt.typeSystem();

        // add method:
        // Object semanticAction(int rule, Object[] sval) {
        //    switch (rule) {
        //        case 0:
        //            // Int ::= LP ID:x RP Int:y { return 0; }
        //            return new Integer(actions$Int$LP$ID$RP$Int(
        //                (Identifier) sval[1],
        //                ((Integer) sval[3]).intValue()));
        //        ...
        //        default:
        //            return null;
        //    }
        // }

        Position pos = Position.COMPILER_GENERATED;

        TypeNode otn = nf.CanonicalTypeNode(pos, ts.Object());
        TypeNode itn = nf.CanonicalTypeNode(pos, ts.Int());
        TypeNode atn = nf.CanonicalTypeNode(pos, ts.arrayOf(pos, ts.Object()));

        ////////////////////////////////////////////////////
        // Introduce locals for the formal params
        ////////////////////////////////////////////////////

        LocalInstance ruleLI = ts.localInstance(pos, Flags.NONE,
                                              ts.Int(), "rule");
        LocalInstance svalLI = ts.localInstance(pos, Flags.NONE,
                                              ts.arrayOf(pos, ts.Object()), "sval");
        Local rule = nf.Local(pos, ruleLI.name());
        rule = rule.localInstance(ruleLI);
        rule = (Local) rule.type(ruleLI.type());

        Local sval = nf.Local(pos, svalLI.name());
        sval = sval.localInstance(svalLI);
        sval = (Local) sval.type(svalLI.type());

        ////////////////////////////////////////////////////
        // Generate the method body
        ////////////////////////////////////////////////////

        Expr null_ = nf.NullLit(pos);
        null_ = null_.type(ts.Null());
        Return retNull = nf.Return(pos, null_);

        List elements = new ArrayList(2*glr.numRules());

        for (int i = 0; i < glr.numRules(); i++) {
            Nonterminal lhs = glr.ruleLhs(i);
            RhsCase rhs = glr.ruleRhs(i);

            if (lhs == null ||rhs == null) {
                continue;
            }

System.out.println("generating code for " + i + ": " + lhs + " ::= " + rhs);

            // case i:
            IntLit label = nf.IntLit(pos, IntLit.INT, i);
            label = (IntLit) label.type(ts.Int());
            Case case_ = nf.Case(pos, label);
            case_ = case_.value(i);
            elements.add(case_);

            List code = new ArrayList();

            // For rule,
            //     lhs ::= rhs0 rhs1:x1 rhs2 rhs3:x3
            //
            // Unmarshal the semantic values x1 and x3:
            //     T1 x1 = unmarshalT1(sval[1]);
            //     T3 x3 = unmarshalT3(sval[3]);
            //
            // Call the action method:
            //     T x = action$lhs$rhs0$rhs1$rhs2$rhs3(x1, x3);
            // or, if void:
            //     action$lhs$rhs0$rhs1$rhs2$rhs3(x1, x3);
            //
            // Marshal and return result:
            //     return marshalT(x);
            // or, if void:
            //     return null;

            String methodName = rhs.actionMethodName(lhs);
            List formalTypes = rhs.actionMethodFormalTypes();

            MethodInstance mi = ts.methodInstance(pos,
                pt, lhs.flags(), lhs.symbolType(), methodName,
                formalTypes, Collections.EMPTY_LIST);

            List args = new ArrayList();
            unmarshalSvals(rhs, sval, code, args, pt);

            if (args.size() != formalTypes.size()) {
                throw new InternalCompilerError("Argument count mismatch between " + mi + " and " + args);
            }

            Special this_ = nf.This(pos);
            this_ = (Special) this_.type(pt);
            Call call = nf.Call(pos, this_, methodName, args);
            call = call.methodInstance(mi);
            call = (Call) call.type(mi.returnType());

            if (mi.returnType().isVoid()) {
                // action(x, y)
                // return null;
                Eval eval = nf.Eval(pos, call);
                code.add(eval);
                code.add(retNull);
            }
            else {
                // return (T) action(x, y);
                Return ret = nf.Return(pos, marshal(call, mi.returnType(), pt));
                code.add(ret);
            }

            Block b = nf.Block(pos, code);
            SwitchBlock sb = nf.SwitchBlock(pos, Collections.singletonList(b));
            elements.add(sb);
        }

        Switch s = nf.Switch(pos, rule, elements);

        List stmts = new ArrayList(2);
        stmts.add(s);
        stmts.add(retNull);

        Block b = nf.Block(pos, stmts);

        ////////////////////////////////////////////////////
        // Generate the method declaration
        ////////////////////////////////////////////////////

        Formal ruleFormal = nf.Formal(pos, Flags.NONE, itn, ruleLI.name());
        ruleFormal = ruleFormal.localInstance(ruleLI);
        Formal svalFormal = nf.Formal(pos, Flags.NONE, atn, svalLI.name());
        svalFormal = svalFormal.localInstance(svalLI);

        List formals = new ArrayList(2);
        formals.add(ruleFormal);
        formals.add(svalFormal);

        MethodInstance mi;

        List formalTypes = new ArrayList(2);
        formalTypes.add(ruleLI.type());
        formalTypes.add(svalLI.type());

        try {
            mi = ts.findMethod(ts.runtimeParserImplType(), "semanticAction",
                               formalTypes, pt);
            mi = mi.flags(Flags.PUBLIC);
            mi = mi.container(pt);
        }
        catch (SemanticException e) {
            throw new InternalCompilerError(e);
        }

        MethodDecl m = nf.MethodDecl(pos, mi.flags(), otn, mi.name(),
                formals, Collections.EMPTY_LIST, b);
        m = m.methodInstance(mi);

        pt.addMethod(mi);

        members(pt).add(m);
    }

    protected void addGlrTableMembers(MutableParserType pt, GLR glr)
        throws SemanticException
    {
        NodeFactory nf = nodeFactory();
        TypeSystem ts = typeSystem();

        Position pos = Position.COMPILER_GENERATED;

        Type t = ts.arrayOf(pos, ts.String());

        String[] s = new String[] { "encodedActionTable",
                                    "encodedOverflowTable",
                                    "encodedGotoTable",
                                    "encodedRuleTable" };

        List tables = new ArrayList();

        tables.add(glr.encodedActionTable());
        tables.add(glr.encodedOverflowTable());
        tables.add(glr.encodedGotoTable());
        tables.add(glr.encodedRuleTable());

        for (int i = 0; i < s.length; i++) {
            FieldInstance fi = ts.fieldInstance(
                pos, pt, Flags.STATIC, t, s[i]);

            // Generate the field declaration:
            // private static String[] table = new String[] { ... };

            List strings = new ArrayList();
            String[] table = (String[]) tables.get(i);

            for (int j = 0; j < table.length; j++) {
                StringLit x = nf.StringLit(pos, table[j]);
                x = (StringLit) x.type(ts.String());
                strings.add(x);
            }

            TypeNode tn;

            tn = nf.CanonicalTypeNode(pos, ts.String());
            ArrayInit init = nf.ArrayInit(pos, strings);
            init = (ArrayInit) init.type(t);
            Expr e = nf.NewArray(pos, tn, Collections.EMPTY_LIST, 1, init);
            e = e.type(t);

            tn = nf.CanonicalTypeNode(pos, t);
            FieldDecl fd = nf.FieldDecl(pos, fi.flags(), tn, s[i], e);
            fd = fd.fieldInstance(fi);

            pt.addField(fi);
            members(pt).add(fd);

            // Generate the block b:
            // { return MyType.table; }
            tn = nf.CanonicalTypeNode(pos, pt);
            Field f = nf.Field(pos, tn, s[i]);
            f = (Field) f.type(t);
            f = f.fieldInstance(fi);

            Return ret = nf.Return(pos, f);
            Block b = nf.Block(pos, Collections.singletonList(ret));

            // Generate the method:
            // protected String[] table() { return MyType.table; }
            MethodInstance mi = ts.methodInstance(
                pos, pt, Flags.PUBLIC, t, s[i],
                Collections.EMPTY_LIST, Collections.EMPTY_LIST);

            tn = nf.CanonicalTypeNode(pos, t);
            MethodDecl m =
                nf.MethodDecl(pos, mi.flags(), tn, mi.name(),
                    Collections.EMPTY_LIST, Collections.EMPTY_LIST, b);
            m = m.methodInstance(mi);

            pt.addMethod(mi);

            members(pt).add(m);
        }
    }

    protected void addGlrEof(MutableParserType pt, GLR glr)
        throws SemanticException
    {
        NodeFactory nf = nodeFactory();
        IbexTypeSystem ts = (IbexTypeSystem) typeSystem();

        Position pos = Position.COMPILER_GENERATED;

        // generate public int eofSymbol() { return N; }

        IntLit N = nf.IntLit(pos, IntLit.INT, glr.eofSymbolNumber());
        N = (IntLit) N.type(ts.Int());

        Return ret = nf.Return(pos, N);
        Block b = nf.Block(pos, Collections.singletonList(ret));

        ClassType parser = ts.runtimeParserImplType();

        MethodInstance mi;

        try {
            mi = ts.findMethod(parser, "eofSymbol", Collections.EMPTY_LIST, pt);
            mi = mi.flags(Flags.PUBLIC);
            mi = mi.container(pt);
        }
        catch (SemanticException e) {
            throw new InternalCompilerError(e);
        }

        TypeNode tn = nf.CanonicalTypeNode(pos, ts.Int());
        MethodDecl m = nf.MethodDecl(pos, mi.flags(), tn, mi.name(),
                            Collections.EMPTY_LIST, Collections.EMPTY_LIST, b);
        m = m.methodInstance(mi);

        pt.addMethod(mi);

        members(pt).add(m);
    }
}
