package polyglot.ext.polyj.types;

import polyglot.types.*;
import polyglot.util.*;

import polyglot.ext.jl.types.*;

import polyglot.ext.polyj.ast.*;
import polyglot.ext.polyj.visit.*;

import java.util.*;
import java.io.Serializable;

/**
 * The goal of a type wrapper is to transform the body of the
 * wrappedType by substituting occurrances of wrappedType.params()
 * with args (stored in the substitutions Map).
 *
 * For each occurence of a parameter, the oldEnclosingType is
 * the innermost class enclosing that occurrence.  The goal is to
 * (lazily) translate the occurence of the parameter in oldEnclosingType
 * into an occurence of the substituted parameter in newEnclosingType.
 *
 * it's always the case that:
 *   t, the current type being scrutinized (could be a parameter
 *      or an array of the parameter type, or an instantiation of
 *      a polymorphic type using the array type)
 * is identical to, or is enclosed by:
 *   oldEnclosingType, the class which is currently being rewritten, which
 * is also identical to, or is enclosed by:
 *   wrappedType.
 *
 * and the goal is to replace t by 
 *   wrap(t, oldEnclosingType, newEnclosingType) such that it
 * is identical to, or is enclosed by:
 *   newEnclsoingType, the reconstruction of oldEnclosingType, using
 *       the substitution map, which
 * is identical to, or is enclosed by:
 *   the InstType which represents the instantiation of
 *   wrappedType instantiated on args.
 */
public class PolyJTypeWrapper implements Serializable {

    /**
     * the type undergoing instantiation
     */
    protected final PolyJClassType wrappedType;

    /**
     * Maps parameters of wrappedType to the argument types on
     * which wrappedType is being instantiated.
     * ParamType => Type
     */
    protected final Map substitutions;

    protected transient PolyJTypeSystem ts;
    
  /**
   * We're doing a lazy reconstruction, and TypeCacheKeys represent
   * enough state to decide whether we need to reconstruct a particular
   * type or not.
   */
    static class TypeCacheKey {
        public Type first;
        public Type second;
        public Type third;
        public TypeCacheKey(Type first, Type second, Type third) {
            this.first = first; this.second = second; this.third = third;
        }
        public boolean equals(Object o) {
            boolean b = 
                o instanceof TypeCacheKey &&
                   (((TypeCacheKey)o).first == null && first == null ||
                    ((TypeCacheKey)o).first.equals(first)) &&
                   (((TypeCacheKey)o).second == null && second == null ||
                    ((TypeCacheKey)o).second.equals(second)) &&
                   (((TypeCacheKey)o).third == null && third == null ||
                    ((TypeCacheKey)o).third.equals(third));
            return b;
        }
        public int hashCode() {
            return (first == null ? 0x80462 : first.hashCode()) ^
                   (second == null ? 0x13579 : second.hashCode()) ^
                   (third == null ? 0xabcde : third.hashCode());
        }

        public String toString() {
            return "(" + first + "," + second + "," + third + ")";
        }
    }

    /**
     * The lazy reconstruction cache.
     * if we're trying to construct newEnclosingType, and we're
     * examining type t, within oldEnclosingType, then
     * the cache will contain a value that represents the result
     * of that work.
     * [t,oldEnclosingType,newEnclosingType]=>result
     */
    protected transient final WeakHashMap/*[TypeCacheKey,Type]*/ cache;

    /**
     * The arguments that wrappedType is being instantiated with.
     */
    protected final List args;

    PolyJTypeWrapper(PolyJClassType wrappedType, List args) {
        this.args = args;
        this.substitutions = makeSubstitutions(wrappedType.params(), args);
        this.wrappedType = wrappedType;
        this.cache = new WeakHashMap();
        this.ts = (PolyJTypeSystem) wrappedType.typeSystem();
    }

    public PolyJClassType wrappedType() { return wrappedType; }

    public List args() { return args; }

    protected static Map makeSubstitutions(List params, List args) {
        HashMap mappings = new HashMap();
        Iterator typeIt = args.iterator();
        Iterator parIt = params.iterator();
        while (typeIt.hasNext() && parIt.hasNext()) {
            mappings.put(parIt.next(), typeIt.next());
        }
        //return Collections.unmodifiableMap(mappings);
        return mappings;
    }


    public Type superType(ClassType r, ClassType newEnclosingType) {
        Type st = r.superType();
        if (st != null && st.isClass())
            return wrap(st, r, newEnclosingType);
        else
            return st;
    }

    public List interfaces(ClassType r, ClassType newEnclosingType) {
        return new CachingTransformingList(r.interfaces(),
                                           new InterfaceTransformation(r,
                                                                       newEnclosingType));
    }

    public List methods(ClassType r,
                        ClassType oldEnclosingType,
                        ClassType newEnclosingType,
                        Map backMap) {
        return new CachingTransformingList(r.methods(),
                                           new MethodTransformation(oldEnclosingType,
                                                                    newEnclosingType,
                                                                    backMap));
    }
    
    public List constructors(ClassType r,
                        ClassType oldEnclosingType,
                        ClassType newEnclosingType,
                        Map backMap) {
        return new CachingTransformingList(r.constructors(), 
                                           new ConstructorTransformation(oldEnclosingType,
                                                                    newEnclosingType,
                                                                    backMap));
    }
    
    public List fields(ClassType r,
                       ClassType oldEnclosingType,
                       ClassType newEnclosingType) {
        return new CachingTransformingList(r.fields(),
                                           new FieldTransformation(oldEnclosingType,
                                                                   newEnclosingType));
    }

    public ClassType memberClassNamed(String name, ClassType r,
                                            ClassType oldEnclosingType,
                                            ClassType newEnclosingType) {
        Type member = r.memberClassNamed(name);
        if (member == null) {
            return null;
        }
        Type wrappedMember = wrap(member, oldEnclosingType, newEnclosingType);
        if (wrappedMember instanceof ClassType &&
            ((ClassType) wrappedMember).isMember()) {
            return (ClassType) wrappedMember;
        } else {
            return null;
        }
    }

    public List memberClasses(ClassType r,
                              ClassType oldEnclosingType,
                              ClassType newEnclosingType) {
        return new CachingTransformingList(r.memberClasses(),
                                           new TypeTransformation(oldEnclosingType,
                                                                  newEnclosingType));
    }
    
    protected Type wrap(Type t, ClassType oldEnclosingType,
                        ClassType newEnclosingType) 
    {
        if (!t.isCanonical()) return t;
        /*
        TypeCacheKey tp = new TypeCacheKey(t, oldEnclosingType, newEnclosingType);
        Type cached = (Type)cache.get(tp);
        if (cached == null) {
            cached = wrapImpl(t, oldEnclosingType, newEnclosingType);
            cache.put(tp, cached);
        }
        */
        return wrapImpl(t, oldEnclosingType, newEnclosingType);
    }

    protected Type wrapImpl(Type t, ClassType oldEnclosingType,
                            ClassType newEnclosingType) {
        if (t == null) return null;
        if (!t.isCanonical()) {
            throw new InternalCompilerError("wrapping a non-canonical type: "+
                t.toString());
        } else if (t.isPrimitive())
            return t;
        else if (t.isArray())
            return t.typeSystem().arrayOf(Position.COMPILER_GENERATED,
                                   wrap(((ArrayType)t).base(),
                                        oldEnclosingType, newEnclosingType));
        else if (t instanceof ParamType) {
            Object o = substitutions.get(t);
            if (o == null)
                return t;
            else
                return (Type)o;
        } else if (t.isClass()) {
            ClassType ct = t.toClass();
            /**
            * if instantiated, see if was instantiated on any of our params,
            *   if so, instantiate on their replacements
            * if inner of wrappedType or an instantiation of an inner of
            *   wrappedType, needs to be wrapped.
            * else, return as is.
            */
            if (ct instanceof InstType) {
                InstType pjit = (InstType)ct;
                pjit = reinstantiate(pjit, oldEnclosingType, newEnclosingType);
                //ct = pjit;
                return pjit;
            }
            
            ClassType nullInstOET = oldEnclosingType;
            if (oldEnclosingType instanceof PolyJClassType
                && !ts.isInstantiated(oldEnclosingType)
                && ts.isInstantiated(ct)) 
            {
                PolyJClassType p = (PolyJClassType) oldEnclosingType;
                if (p.isPolymorphic()) {
                    nullInstOET = ts.nullInstantiate(p);
                }
            }
            
            if (ts.equals(ct, nullInstOET) ||
                ts.isEnclosed(ct, nullInstOET)) {
                return wrapClassType(ct, oldEnclosingType, newEnclosingType);
            }

            if (ct.isInner()) {
                ClassType cc = ct.outer().toClass();
                if (cc instanceof InstType || cc.isInner()) {
                    Type newCC = wrap(cc, oldEnclosingType, newEnclosingType);
                    if (! ts.equals(newCC, cc)) {
                        // NJN: Where is the check that ct is a _member_?
                        return newCC.toClass().memberClassNamed(
                            PolyJTranslator.nameForClass(ct)).toClass();
                    }
                }
                return ct;
            } else {
                return ct;
            }
        } else {
            throw new InternalCompilerError("wrapping a canonical non-class, "
                + "non-array, non-parameter type: " + t.toString());
        }
    }

    protected InstType reinstantiate(InstType pjit,
                                     ClassType oldEnclosingType,
                                     ClassType newEnclosingType) 
    {
        if (!pjit.typeSystem().equals(pjit.instType().toClass(),
                                      wrappedType.toClass()) &&
            pjit.isNullInstantiation()) /*easy*/
            return pjit;
        boolean substitutionOccurred = false;
        TypedList substitutedArgs =
                                   new TypedList(new ArrayList(pjit.actualArgs().size()),
                                                 Type.class, false);
        for (Iterator it = pjit.actualArgs().iterator(); it.hasNext(); ) {
            Type arg = (Type)it.next();
            Object res = substitutions.get(arg);
            if (res != null) {
                substitutionOccurred = true;
                substitutedArgs.add(res);
            } else {
                Type wrappedArg = wrap(arg, oldEnclosingType, newEnclosingType);
                if (wrappedArg != /*pointer equality*/ arg) {
                    substitutionOccurred = true;
                    arg = wrappedArg;
                }
                substitutedArgs.add(arg);
            }
        }

        PolyJClassType instType = pjit.instType();
	// FIXME:  Is this code correct?  It causes the reinstantiation
	// to abort before the new substitutions are made.  See the log
	// message for more details, and tests/bad/semant/BadInnerInst.pj
	// for a test case.  MRC
        //if (instType instanceof WrappedClassType)
        //    return pjit;
        PolyJClassType wrappedInst =
                                    (PolyJClassType) wrap(instType.toClass(),
                                                          oldEnclosingType,
                                                          newEnclosingType);
        substitutionOccurred |= (instType != /*pointer equality*/ wrappedInst);
        if (substitutionOccurred) {
            PolyJTypeSystem pjTS = 
                                  (PolyJTypeSystem)instType.toClass().typeSystem();
            return pjTS.uncheckedInstantiate(wrappedInst, substitutedArgs);
        } else 
            return pjit;
    }
    
    protected MethodInstance wrapMethod(MethodInstance mi,
                                        ClassType oldEnclosingType,
                                        ClassType newEnclosingType,
                                        Map backMap) {
        List wrappedArgs = new TransformingList(mi.formalTypes(),
            new TypeTransformation(oldEnclosingType, newEnclosingType));
        wrappedArgs = new TypedList(wrappedArgs, Type.class, true);
        
        List wrappedExns = new TransformingList(mi.throwTypes(),
            new TypeTransformation(oldEnclosingType,  newEnclosingType));
        wrappedExns = new TypedList(wrappedExns, Type.class, true);

        MethodInstance newMI = wrappedType.typeSystem().methodInstance(
            mi.position(), newEnclosingType, mi.flags(),
            wrap(mi.returnType(), oldEnclosingType, newEnclosingType),
            mi.name(),    wrappedArgs, wrappedExns);
        
        if (backMap != null) {
            backMap.put(newMI, mi);
        }
        
        return newMI;
    }
    
    protected ConstructorInstance wrapConstructor(ConstructorInstance ci,
                                        ClassType oldEnclosingType,
                                        ClassType newEnclosingType,
                                        Map backMap) {
        List wrappedArgs = new TransformingList(ci.formalTypes(),
            new TypeTransformation(oldEnclosingType, newEnclosingType));
        wrappedArgs = new TypedList(wrappedArgs, Type.class, true);
        
        List wrappedExns = new TransformingList(ci.throwTypes(),
            new TypeTransformation(oldEnclosingType,  newEnclosingType));
        wrappedExns = new TypedList(wrappedExns, Type.class, true);

        ConstructorInstance newCI = wrappedType.typeSystem().constructorInstance(
            ci.position(), newEnclosingType, ci.flags(),
            wrappedArgs, wrappedExns);
        
        if (backMap != null) {
            backMap.put(newCI, ci);
        }
        
        return newCI;
    }

    protected FieldInstance wrapField(FieldInstance fi, ClassType oldEnclosingType, ClassType newEnclosingType) {
        FieldInstance rv = newEnclosingType.typeSystem().fieldInstance(fi.position(),
            newEnclosingType, fi.flags(), wrap(fi.type(), oldEnclosingType, newEnclosingType), fi.name());
        if (fi.isConstant()) {
            rv.constantValue(fi.constantValue());
        }
        
        return rv;
    }

    
    class TypeTransformation implements Transformation {
        ClassType oldEnclosingType;
        ClassType newEnclosingType;
        
        public TypeTransformation(ClassType oldEnclosingType, ClassType newEnclosingType) {
            this.oldEnclosingType = oldEnclosingType;
            this.newEnclosingType = newEnclosingType;
        }
        
        public Object transform(Object o) {
            return PolyJTypeWrapper.this.wrap((Type)o, oldEnclosingType, newEnclosingType);
        }
    }

    class MethodTransformation implements Transformation {
        ClassType oldEnclosingType;
        ClassType newEnclosingType;
        Map backMap;
        
        public MethodTransformation(ClassType oldEnclosingType,    ClassType newEnclosingType, Map backMap) {
            this.oldEnclosingType = oldEnclosingType;
            this.newEnclosingType = newEnclosingType;
            this.backMap = backMap;
        }
        
        public Object transform(Object o) {
            return PolyJTypeWrapper.this.wrapMethod((MethodInstance)o, oldEnclosingType, newEnclosingType, backMap);
        }
    }

    class ConstructorTransformation implements Transformation {
        ClassType oldEnclosingType;
        ClassType newEnclosingType;
        Map backMap;
        
        public ConstructorTransformation(ClassType oldEnclosingType, ClassType newEnclosingType, Map backMap) {
            this.oldEnclosingType = oldEnclosingType;
            this.newEnclosingType = newEnclosingType;
            this.backMap = backMap;
        }
        
        public Object transform(Object o) {
            return PolyJTypeWrapper.this.wrapConstructor((ConstructorInstance)o, oldEnclosingType, newEnclosingType, backMap);
        }
    }
    
    class FieldTransformation implements Transformation {
        ClassType oldEnclosingType;
        ClassType newEnclosingType;
        
        public FieldTransformation(ClassType oldEnclosingType, ClassType newEnclosingType) {
            this.oldEnclosingType = oldEnclosingType;
            this.newEnclosingType = newEnclosingType;
        }
        
        public Object transform(Object o) {
            return PolyJTypeWrapper.this.wrapField((FieldInstance)o, oldEnclosingType, newEnclosingType);
        }
    }

    class InterfaceTransformation implements Transformation {
        ClassType r;
        ClassType newEnclosingType;
        
        public InterfaceTransformation(ClassType r, ClassType newEnclosingType) {
            this.r = r;
            this.newEnclosingType = newEnclosingType;
        }

        public Object transform(Object o) {
            return wrap((Type)o,r, newEnclosingType);
        }
    }

    ClassType wrapClassType(ClassType r, ClassType oldEnclosingType, ClassType newEnclosingType) {
        if (r instanceof PolyJClassType) {
            PolyJClassType p = (PolyJClassType) r;
            return ts.WrappedClassType(p, oldEnclosingType, newEnclosingType, this);
        } else {
            throw new InternalCompilerError("Unable to wrap a non-PolyJ class");
        }
    }
    
}
