package polyglot.ext.polyj.visit;

import polyglot.ast.*;
import polyglot.frontend.*;
import polyglot.types.*;
import polyglot.util.*;
import polyglot.visit.*;

import polyglot.ext.polyj.ast.*;
import polyglot.ext.polyj.types.*;
import polyglot.ext.polyj.util.*;

import java.util.*;

/**
 * Generates the "secondary" instantiation bindings in the
 * PolyJ translation table that will be necessary in
 * order to generate adapter interfaces and classes.
 * These are the instantiations that occur indirectly
 * through partial instantiations in the source code.
 *
 * All adapter bindings for all sources 
 * must have been generated before running this visitor.
 *
 * This doesn't really need to be a visitor, because there's
 * nothing in the AST that needs visited.  It just needs
 * to be a pass of some sort.
 */
public class InstBindingsGenerator extends NodeVisitor
{
    PolyJTypeSystem pjts;
    TranslationTable tt;

    public InstBindingsGenerator(PolyJTypeSystem ts) {
        pjts = ts;
        tt = pjts.translationTable();	
    }

    public Node override(Node n) {
        // Don't visit anything.  FIXME: this should be a pass, not a visitor.
        return n;
    }

    public void finish() {
	Set s = new HashSet();
        for (Iterator it = tt.instantiations().iterator(); it.hasNext(); ) {
            ClassType t = (ClassType) it.next();

	    // Although we may not have recorded any instantiations of the
	    // supertypes of "t", we still have to find any indirect
	    // instantiations in them.  This is because "t" may inherit
	    // code from them that references adapter classes for those
	    // instantiations.
	    while (!t.equals(pjts.Object())) {
		s.addAll(indirectInst(t));
		t = t.superType().toClass();
	    }
        }

        // We have to delay adding all the new instantiations until
        // after we've iterated through all the old ones, otherwise
        // we'll get a ConcurrentModificationException from
        // the Iterator above.
        pjts.addInstantiations(s);

    }

    /**
     * Return all the indirect instantiations caused by <code>t</code>.
     */
    protected Set indirectInst(ClassType t) {
	Set s = new HashSet();
	
        InstantiationBinding ib = pjts.InstantiationBinding(t);
        if (ib.needsAdapterClass()) {
            Iterator adapterBindings = tt.getAdapterBindings(
                pjts.uninstantiate(ib.type()));

            while (adapterBindings.hasNext()) {
                AdapterBinding adapter
		    = (AdapterBinding) adapterBindings.next();

                ClassType inst = tt.getInstantiatedClass(
                    ib.type(), adapter);
                //System.out.println("Indirect inst: " + inst);
		    
		if (!tt.instantiations().contains(inst)) {
	 	    s.add(inst);
		}
            }
        }

	return s;
    }
    
}
